from fisher_py.data.raw_file_thread_accessor import RawFileThreadAccessor
from fisher_py.net_wrapping import ThermoFisher
from fisher_py.raw_file_reader.raw_file_access import RawFileAccess


class RawFileReaderFactory:

    @staticmethod
    def read_file(filename: str) -> RawFileAccess:
        """        
         Summary:
             Open a raw file for reading.
        
         Parameters:
           fileName:
             Name of file to read
        
         Returns:
             Access to the contents of the file.
        """
        assert type(filename) is str
        return RawFileAccess._get_wrapper_(ThermoFisher.CommonCore.Data.Business.RawFileReaderFactory.ReadFile(filename))

    @staticmethod
    def create_thread_manager(filename: str) -> RawFileThreadAccessor:
        """
         Summary:
             Open a raw file for reading, creating a manager tool, such that multiple threads
             can access the same open file.
        
         Parameters:
           fileName:
             Name of file to read
        
         Returns:
             Access to the contents of the file.
        """
        assert type(filename) is str
        return RawFileThreadAccessor._get_wrapper_(ThermoFisher.CommonCore.Data.Business.RawFileReaderFactory.CreateThreadManager(filename))



import pytest
from fisher_py.data.device import Device
from fisher_py.exceptions import RawFileException
from fisher_py.exceptions.raw_file_exception import NoSelectedDeviceException, NoSelectedMsDeviceException
from fisher_py.raw_file_reader import RawFileAccess
from tests import assert_attributes, path_for

REFERENCE_RAW_FILE = 'Angiotensin_325-CID.raw'

def test_raw_file_import_fails_for_invalid_path():
    try:
        RawFileAccess('does-not-exist.raw')
        assert False, 'FileNotFoundError should have been thrown'
    except FileNotFoundError:
        pass

def test_raw_file_import_fails_for_invalid_file():
    try:
        RawFileAccess(path_for('invalid.raw'))
        assert False, 'RawFileException should have been thrown'
    except RawFileException:
        pass

def test_raw_file_import_succeeds_with_valid_file():
    access = RawFileAccess(path_for(REFERENCE_RAW_FILE))
    assert access.file_error.has_error == False, 'Import should succeed'

@pytest.mark.parametrize('attribute', ['run_header', 'run_header_ex', 'scan_events', 'status_log_plottable_data'])
def test_access_to_instrument_dependent_attributes_fails_without_instrument_selection(attribute: str):
    access = RawFileAccess(path_for(REFERENCE_RAW_FILE))
    try:
        getattr(access, attribute)
        assert False, 'Access should fail'
    except (NoSelectedDeviceException, NoSelectedMsDeviceException):
        pass

@pytest.mark.parametrize('attribute', ['file_header', 'sample_information'])
def test_attribute_available_without_instrument_selection_is_as_expected(attribute: str):
    access = RawFileAccess(path_for(REFERENCE_RAW_FILE))
    assert_attributes(access, attribute)

def test_device_selection_works():
    access = RawFileAccess(path_for(REFERENCE_RAW_FILE))
    assert access.selected_instrument.device_type == Device.none

    access.select_instrument(Device.MS, 1)
    assert access.selected_instrument.device_type == Device.MS
    assert access.selected_instrument.instrument_index == 1

@pytest.mark.parametrize('attribute', ['run_header', 'run_header_ex'])
def test_instrument_dependent_attribute_matches_expected_data(attribute: str):
    access = RawFileAccess(path_for(REFERENCE_RAW_FILE))
    access.select_instrument(Device.MS, 1)
    assert_attributes(access, attribute)
    


import datetime
from enum import Enum
from pathlib import Path
import inspect
import requests
import os
import json

TEST_DIR = os.path.dirname(__file__)

def download_if_required(path: str):
    """Downloads a file if not present in the filesystem"""
    # do nothing if file already exists
    if os.path.isfile(path):
        return

    # check if there exists a .ftp file to fetch the target file    
    file_path = Path(path)
    ftp_file_path = file_path.with_name(f'{os.path.splitext(file_path.name)[0]}.ftp')

    if not os.path.isfile(ftp_file_path):
        raise FileNotFoundError(f'Unable to find either file or download source for "{path}".')
    
    # read the .ftp file
    with open(ftp_file_path, 'r', encoding='utf-8') as f:
        url = f.read().replace('\n', '').strip()
    
    # try to download missing file
    result = requests.get(url)
    if result.status_code != 200:
        raise Exception(f'Unable to download "{path}" from "{url}".')
    
    with open(path, 'wb') as f:
        f.write(result.content)


def path_for(file: str) -> str:
    """Returns path to test data"""
    path = os.path.join(TEST_DIR, 'data', file)
    download_if_required(path)
    return path

def info_json_for(file: str) -> str:
    name = os.path.splitext(os.path.split(file)[-1])[0]
    return path_for(f'{name}.info.json')


def assert_attribute(actual, expected):
    """Asserts that the given attribute matches the expected value"""
    if type(expected) is dict:
        keys = expected.keys()
        if len(keys) == 2 and 'type' in keys and 'value' in keys:
            type_name = expected['type']
            assert type(actual).__name__ == type_name, f'Expected value to have "{type_name}" type'

            value = expected['value']
            if type_name == 'datetime':
                value = datetime.datetime.fromisoformat(value)
            elif isinstance(actual, Enum):
                actual = actual.value

            assert actual == value, f'Expected value to be "{value}"'
        else:
            for attribute_name, attribute_value in expected.items():
                assert hasattr(actual, attribute_name), f'Expected attribute "{attribute_name}"'
                assert_attribute(getattr(actual, attribute_name), attribute_value)
    elif isinstance(expected, list):
        assert isinstance(actual, list), 'Expected value to be list'
        assert len(actual) == len(expected), f'Expected list value to have length {len(expected)}'

        for expected_item, actual_item in zip(expected, actual):
            assert_attribute(actual_item, expected_item)
    elif type(expected) is float:
        assert type(actual) is float, f'Expected value to be float'
        assert abs(actual - expected) < 1e-9
    else:
        assert actual == expected, f'Expected value to be "{expected}"'

def assert_attributes(access, path: str):
    """Asserts object attributes against *.info.json file."""

    # load info json for .raw file
    info_json_path = info_json_for(access.file_name)

    if not os.path.isfile(info_json_path):
        raise FileNotFoundError(f'The info.json file "{info_json_path}" does not exist.')
    
    with open(info_json_path, 'r', encoding='utf-8') as f:
        info_json = json.load(f)

    # locate node of interest in info json
    path_parts = path.split('.') if '.' in path else [path]
    expected_node = info_json
    actual_node = access
    for attr in path_parts:
        expected_node = expected_node[attr]
        actual_node = getattr(actual_node, attr)

    assert_attribute(actual_node, expected_node)

def capture_attribute_as_dict(access, path: str):
    """Capture object attribute and output to dictionary"""
    
    # locate node of interest in info json
    path_parts = path.split('.') if '.' in path else [path]
    node = access
    for attr in path_parts:
        node = getattr(node, attr)
    
    # generate dictionary
    capture = dict()
    for attribute_name, attribute_value in inspect.getmembers(node):
        if attribute_name.startswith('_'):
            continue
        
        if type(attribute_value) is dict:
            capture[attribute_name] = capture_attribute_as_dict(attribute_value)
        elif type(attribute_value) is list:
            list_value = list()
            for item in attribute_value:
                if type(item) is dict:
                    list_value.append(capture_attribute_as_dict(item))
                else:
                    list_value.append(item)
            capture[attribute_name] = list_value
        elif isinstance(attribute_value, Enum):
            capture[attribute_name] = { 'type': type(attribute_value).__name__, 'value': attribute_value.value }
        elif type(attribute_value) is datetime.datetime:
            capture[attribute_name] = { 'type': 'datetime', 'value': str(attribute_value) }
        elif type(attribute_value).__name__ == 'method':
            continue
        else:
            capture[attribute_name] = attribute_value
    
    return capture

def capture_attribute(access, path: str):
    """Capture object attribute and output to console"""
    capture = capture_attribute_as_dict(access, path)
    print(json.dumps(capture, indent=4))
    



from fisher_py.net_wrapping.wrapped_net_list import WrappedNetList
from fisher_py.utils import to_net_list
import json

EPS = 1e-16

def test_wrapped_lists_can_be_created():
    net_list = to_net_list([1, 2, 3], int)
    wrapped_list = WrappedNetList[int](net_list)
    assert len(wrapped_list) == 3    

def test_wrapped_lists_support_floats():
    values = [6.4, 7.7, 34.23, 1e-9, 13e12]
    net_list = to_net_list(values, float)
    wrapped_list = WrappedNetList[float](net_list)

    for actual, expected in zip(values, wrapped_list):
        assert abs(actual - expected) < EPS, f'Expected value "{expected}" but actual value was "{actual}"'

def test_wrapped_lists_support_strings():
    values = ['This is a string', 'another string', 'still more strings']
    net_list = to_net_list(values, str)
    wrapped_list = WrappedNetList[str](net_list)

    for actual, expected in zip(values, wrapped_list):
        assert actual == expected, f'Expected value "{expected}" but actual value was "{actual}"'

def test_wrapped_lists_can_be_serialized():
    values = [3, 6, 8, 2]
    net_list = to_net_list(values, int)
    wrapped_list = WrappedNetList[int](net_list)
    json_str = json.dumps(wrapped_list)
    assert json_str == '[3, 6, 8, 2]'

def test_wrapped_lists_items_can_be_added():
    values = [3, 6, 8, 2]
    net_list = to_net_list(values, int)
    wrapped_list = WrappedNetList[int](net_list)
    wrapped_list.append(77)
    wrapped_list.append(80)
    assert wrapped_list[-2] == 77
    assert wrapped_list[-1] == 80

def test_wrapped_lists_items_can_be_inserted():
    values = [3, 6, 8, 2]
    net_list = to_net_list(values, int)
    wrapped_list = WrappedNetList[int](net_list)
    wrapped_list.insert(1, 10)
    wrapped_list.insert(-2, 7)
    assert wrapped_list[1] == 10
    assert wrapped_list[3] == 7

def test_wrapped_lists_items_can_be_removed():
    values = [3, 6, 8, 2]
    net_list = to_net_list(values, int)
    wrapped_list = WrappedNetList[int](net_list)
    wrapped_list.remove(8)
    wrapped_list.remove(3)
    assert len(wrapped_list) == len(values) - 2
    assert 8 not in wrapped_list
    assert 3 not in wrapped_list

def test_wrapped_lists_can_be_cleared():
    values = [3, 6, 8, 2]
    net_list = to_net_list(values, int)
    wrapped_list = WrappedNetList[int](net_list)
    wrapped_list.clear()
    assert len(wrapped_list) == 0

def test_wrapped_lists_can_use_python_indexing():
    values = [3, 6, 8, 2]
    net_list = to_net_list(values, int)
    wrapped_list = WrappedNetList[int](net_list)
    result = wrapped_list[1:4]
    assert result[0] == 6
    assert result[1] == 8
    assert result[2] == 2




from fisher_py.raw_file_reader.raw_file_reader_adapter import RawFileReaderAdapter
from tests import path_for


def test_raw_file_adapter_can_load_file():
    file = RawFileReaderAdapter.file_factory(path_for('Angiotensin_325-CID.raw'))


class RawFileException(Exception):

    def __init__(self, message: str):
        super().__init__(message)


class NoSelectedDeviceException(Exception):
    pass


class NoSelectedMsDeviceException(Exception):

    def __init__(self, message: str):
        super().__init__(message)



from fisher_py.raw_file_reader import RawFileAccess


class RawFileReaderAdapter(object):
    """
    Utility to load raw files
    """

    @staticmethod
    def file_factory(raw_file: str):
        """
        Create an IRawDataExtended interface to read data from a raw file

        :param raw_file: Path to raw file
        :return: Returns raw file
        """
        return RawFileAccess(raw_file)



from typing import TypeVar, Generic, Union
import clr

clr.AddReference('System')
import System
import System.Collections.Generic as generic

T = TypeVar('T')

class WrappedNetArray(Generic[T], list):

    def __getitem__(self, index: Union[int, slice]) -> T:
        if type(index) is slice:
            start, step, stop = index.start, index.step, index.stop
            if step is None:
                step = 1
            start, step, stop = map(self.__ensure_positive_index__, [start, step, stop])
            sublist = []
            for i in range(start, stop, step):
                sublist.append(self.__wrapped_iterable[i])            
            return sublist
        index = self.__ensure_positive_index__(index)
        return self.__wrapped_iterable[index]
    
    def __setitem__(self, index: int, value: T):
        index = self.__ensure_positive_index__(index)
        super().__setitem__(index, value)
        self.__wrapped_iterable[index] = value

    def append(self, obj):
        raise NotImplementedError('Appending not allowed on arrays.')
    
    def insert(self, index, obj):
        raise NotImplementedError('Appending not allowed on arrays.')
    
    def remove(self, value):
        raise NotImplementedError('Removing not allowed on arrays.')
    
    def clear(self):
        raise NotImplementedError('Removing not allowed on arrays.')

    def __init__(self, net_iterable):
        super().__init__([i for i in net_iterable])
        self.__wrapped_iterable = net_iterable
    
    def __ensure_positive_index__(self, index: int) -> int:
        if index < 0:
            index = len(self.__wrapped_iterable) + index
        return index




from fisher_py.exceptions.raw_file_exception import RawFileException
from fisher_py.raw_file import RawFile
from tests import path_for

TEST_FILE = 'Angiotensin_325-CID.raw'
PRECURSOR_MZ = 325
TOLERANCE = 1e-6
TIMES = [0.00213759, 0.00539962, 0.00867493, 0.01190376, 0.01514103, 0.0183659 , 0.02165696, 0.02493226, 0.02813163, 0.0314831]

def test_raw_file_initialization_fails_on_nonexisting_path():
    try:
        RawFile('non-existent.raw')
        assert False, 'Expected error when trying to load nonexistent file'
    except FileNotFoundError:
        pass

def test_raw_file_initialization_fails_on_invalid_file():
    try:
        RawFile(path_for('invalid.raw'))
        assert False, 'Expected error when trying to load invalid file'
    except RawFileException:
        pass

def test_raw_file_initialization_succeeds_on_valid_file():
    RawFile(path_for(TEST_FILE))

def test_raw_file_number_of_scans_matches_expected_value():
    file = RawFile(path_for(TEST_FILE))
    assert file.number_of_scans == 10

def test_raw_file_first_scan_matches_expected_value():
    file = RawFile(path_for(TEST_FILE))
    assert file.first_scan == 1

def test_raw_file_last_scan_matches_expected_value():
    file = RawFile(path_for(TEST_FILE))
    assert file.last_scan == 10

def test_raw_file_total_time_min_matches_expected_value():
    file = RawFile(path_for(TEST_FILE))
    assert abs(file.total_time_min - 0.031483095733333334) < TOLERANCE

def test_raw_file_ms2_filter_masses_matches_expected_value():
    file = RawFile(path_for(TEST_FILE))
    assert len(file.ms2_filter_masses) == 1
    assert abs(file.ms2_filter_masses[0] - PRECURSOR_MZ) < TOLERANCE

def test_raw_file_chromatogram_can_be_retrieved():
    file = RawFile(path_for(TEST_FILE))
    t, c = file.get_chromatogram(PRECURSOR_MZ, TOLERANCE)
    assert len(t) == 0
    assert len(c) == 0

def test_raw_file_tic_ms2_can_be_retrieved():
    file = RawFile(path_for(TEST_FILE))
    times, total_ion_current = file.get_tic_ms2(PRECURSOR_MZ, TOLERANCE)

    expected_ion_currents = [37451832.53613281, 37663065.91699219, 45855594.28125, 30816050.90820312, 41055303.14746094, 34280428.34375, 35895350.90429688, 33482430.11523438, 36745244.81738281, 35045496.85644531]
    for t_actual, t_expected, tic_actual, tic_expected in zip(times, TIMES, total_ion_current, expected_ion_currents):
        assert abs(t_actual - t_expected) < TOLERANCE
        assert abs(tic_actual - tic_expected) < TOLERANCE

def test_raw_file_scan_ms2_can_be_retrieved():
    file = RawFile(path_for(TEST_FILE))
    lenths = [175, 199, 163, 150, 142, 186, 174, 114, 203]

    for rt, l in zip(TIMES, lenths):
        masses, charges, intensities, _ = file.get_scan_ms2(rt)
        assert len(masses) == l
        assert len(charges) == l
        assert len(intensities) == l

def test_raw_file_getting_scan_from_number_works_as_expected():
    file = RawFile(path_for(TEST_FILE))
    lenths = [175, 199, 163, 150, 142, 186, 174, 114, 203]

    for i, l in zip(range(1, 11), lenths):
        masses, charges, intensities, _ = file.get_scan_from_scan_number(i)
        assert len(masses) == l
        assert len(charges) == l
        assert len(intensities) == l

def test_raw_file_getting_scan_number_from_retention_time_works_as_expected():
    file = RawFile(path_for(TEST_FILE))

    for i, rt in enumerate(TIMES[:-1]):
        sn = file.get_scan_number_from_retention_time(rt)
        assert sn == i + 1

def test_raw_file_ms2_getting_scan_number_from_retention_time_works_as_expected():
    file = RawFile(path_for(TEST_FILE))

    for i, rt in enumerate(TIMES[:-1]):
        sn, _ = file.get_ms2_scan_number_from_retention_time(rt)
        assert sn == i + 1

def test_raw_file_get_scan_event_str_from_scan_number_works_as_expected():
    file = RawFile(path_for(TEST_FILE))
    desc = file.get_scan_event_str_from_scan_number(1)
    assert desc == 'FTMS + p ESI Full ms2 325.0000@cid35.00 [150.0000-2000.0000]'


from fisher_py.net_wrapping.net_wrapper_base import NetWrapperBase
from fisher_py.raw_file_reader.raw_file_access import RawFileAccess


class RawFileThreadAccessor(NetWrapperBase):
    """    
     Summary:
         Defines an object which can create accessors for multiple threads to access the
         same raw data.
    """

    def create_thread_accessor(self) -> RawFileAccess:
        """
         Summary:
             This interface method creates a thread safe access to raw data, for use by a
             single thread. Each time a new thread (async call etc.) is made for accessing
             raw data, this method must be used to create a private object for that thread
             to use. This interface does not require that the application performs any locking.
             In some implementations this may have internal locking (such as when based on
             a real time file, which is continually changing in size), and in some implementations
             it may be lockless.
        
         Returns:
             An interface which can be used by a thread to access raw data
        """
        return RawFileAccess._get_wrapper_(self._wrapped_object.CreateThreadAccessor())
